/*
 *
 *  VSTRING Library supporting structures and functions
 *  (c) Vladi Belperchinov-Shabanski "Cade" <cade@biscom.net> 1998-2000
 *  Distributed under the GPL license, you should receive copy of GPL!
 *
 * $Id: vstrlib.cpp,v 1.5 2002/04/26 06:40:08 cade Exp $
 *
 */

#include "vstrlib.h"

/***************************************************************************
**
** VARRAY
**
****************************************************************************/

  VArray::VArray()
  {
    _data = NULL;
    _size = 0;
    _count = 0;
    _max_len = 0;
    _min_len = (unsigned int)-1; /* should underflow to max value :) */
    _ret_str = (char*)NULL;
  }

  VArray::VArray( const VArray& arr )
  {
    _data = NULL;
    _size = 0;
    _count = 0;
    _max_len = 0;
    _min_len = (unsigned int)-1; /* should underflow to max value :) */
    _ret_str = (char*)NULL;
    *this = arr;
  }

  VArray::VArray( const VTrie& tr )
  {
    _data = NULL;
    _size = 0;
    _count = 0;
    _max_len = 0;
    _min_len = (unsigned int)-1; /* should underflow to max value :) */
    _ret_str = (char*)NULL;
    *this = tr;
  }

  VArray::~VArray()
  {
    zap();
  }

  void VArray::resize( int new_size )
  {
    ASSERT( new_size >= 0 );
    if ( new_size == 0 )
      {
      delete _data;
      _data = NULL;
      _size = 0;
      _count = 0;
      return;
      }
    new_size  = new_size / VARRAY_BLOCK_SIZE + (new_size % VARRAY_BLOCK_SIZE != 0);
    new_size *= VARRAY_BLOCK_SIZE;
    if ( new_size == _size ) return;
    String** new_data = new String*[ new_size ];
    ASSERT( new_data );
    memset( new_data, 0, new_size * sizeof(String*) );
    if ( _data )
      {
      memcpy( new_data, _data,
              (_size < new_size ? _size : new_size) * sizeof(String*) );
      delete _data;
      }
    _size = new_size;
    _data = new_data;
  }

  void VArray::ins( int n, const char* s )
  {
    ASSERT( n >= 0 && n <= _count );
    if ( _count == _size ) resize( _size + 1 );
    memmove( &_data[0] + n + 1,
             &_data[0] + n,
             ( _count - n ) * sizeof(String*) );
    _count++;

    _data[n] = new String;
    _data[n]->compact = 1;
    _data[n]->set( s );

    int sl = str_len( *_data[n] );
    if ( sl < _min_len )  _min_len = sl;
    if ( sl > _max_len )  _max_len = sl;
  }

  void VArray::del( int n )
  {
    if ( n < 0 || n >= _count ) return;
    delete _data[n];
    memmove( &_data[0] + n,
             &_data[0] + n + 1,
             ( _count - n ) * sizeof(String*) );
    _count--;
    if ( _size - _count > VARRAY_BLOCK_SIZE ) resize( _count );
  }

  void VArray::set( int n, const char* s )
  {
    ASSERT( n >= 0 );
    if ( n >= _count )
      {
      int i = n - _count + 1;
      while ( i-- ) push( "" );
      }
    ASSERT( n < _count );
    _data[n]->set( s );
  }

  const char* VArray::get( int n )
  {
    if ( n < 0 || n >= _count )
      return NULL;
    else
      return _data[n]->data();
  }

  void VArray::zap()
  {
    if ( _count )
      for( int z = 0; z < _count; z++ )
        if ( _data[z] )
          delete _data[z];
    resize( 0 );

    _data = NULL;
    _size = 0;
    _count = 0;
    _max_len = 0;
    _min_len = (unsigned int)-1; /* should underflow to max value :) */
    _ret_str = "";
  }

  int VArray::push( const char* s )
  {
    ins( _count, s );
    return _count;
  }

  const char* VArray::pop()
  {
    if ( _count == 0 ) return NULL;
    _ret_str = get( _count - 1 );
    del( _count - 1 );
    return _ret_str.data();
  }

  int VArray::unshift( const char* s )
  {
    ins( 0, s );
    return _count;
  }

  const char* VArray::shift()
  {
    if ( _count == 0 ) return NULL;
    _ret_str = get( 0 );
    del( 0 );
    return _ret_str.data();
  }

  int VArray::merge( VTrie *tr )
  {
    VArray va;
    int cnt = tr->keys( &va );
    for( int z = 0; z < cnt; z++ )
      {
      const char* key = va[z];
      push( key );
      push( tr->get( key ) );
      }
    return _count;
  };

  int VArray::merge( VArray *arr )
  {
    int cnt = arr->count();
    for( int z = 0; z < cnt; z++ )
      push( arr->get( z ) );
    return _count;
  };

  int VArray::fload( const char* fname )
  {
    FILE* f = fopen( fname, "rt" );
    if (!f) return 1;
    int r = fload( f );
    fclose(f);
    return r;
  };

  int VArray::fsave( const char* fname )
  {
    FILE* f = fopen( fname, "wt" );
    if (!f) return 1;
    int r = fsave( f );
    fclose(f);
    return r;
  };

  int VArray::fload( FILE* f )
  {
    char buf[1024];
    String str;
    while( fgets( buf, sizeof(buf)-1, f ) )
      {
      str += buf;
      if ( str_get_ch( str, -1 ) != '\n' ) continue;
      while ( str_get_ch( str, -1 ) == '\n' ) str_trim_right( str, 1 );
      push( str );
      str = "";
      }
    return 0;
  };

  int VArray::fsave( FILE* f )
  {
    for( int z = 0; z < _count; z++ )
      {
      size_t len = strlen( get(z) );
      if ( fwrite( get(z), 1, len, f ) != len ) return 2;
      if ( fwrite( "\n", 1, 1, f ) != 1 ) return 2;
      }
    return 0;
  };

  void VArray::sort()
  {
    if ( count() > 1 )
      q_sort( 0, count() - 1 );
  };

  void VArray::q_sort( int lo, int hi )
  {
    int m, l, r;
    const char* v;

    m = ( hi + lo ) / 2;
    v = _data[m]->data();
    l = lo;
    r = hi;

    do
      {
      while( (l <= hi) && (strcmp(_data[l]->data(),v) < 0) ) l++;
      while( (r >= lo) && (strcmp(v,_data[r]->data()) < 0) ) r--;
      if ( l <= r )
        {
        String *t;
        t = _data[l];
        _data[l] = _data[r];
        _data[r] = t;
        l++;
        r--;
        }
      }
    while( l <= r );

    if ( lo < r ) q_sort( lo, r );
    if ( l < hi ) q_sort( l, hi );
  };

  void VArray::reverse()
  {
    int m = _count / 2;
    for( int z = 0; z < m; z++ )
      {
      String *t;
      t = _data[z];
      _data[z] = _data[_count-1-z];
      _data[_count-1-z] = t;
      }
  };

  void VArray::shuffle() /* Fisher-Yates shuffle */
  {
    int i = _count - 1;
    while( i >= 0 )
      {
      int j = rand() % ( i + 1 );
      String *t;
      t = _data[i];
      _data[i] = _data[j];
      _data[j] = t;
      i--;
      }
  };

  void VArray::split( const char* res, const char* str, int maxcount )
  {
    zap();
    
    regexp *re = regcomp( res );
    ASSERT( re );

    const char* ps = str;

    String s;
    while( regexec( re, ps ) )
      {
      if ( maxcount != -1 )
        {
        maxcount--;
        if ( maxcount == 0 ) break;
        }
      int l = re->startp[0] - ps;
      s.setn( ps, l );
      push( s );
      ps = re->endp[0];
      }
    push( ps );

    free( re );
  };

  void VArray::split_str( const char* res, const char* str, int maxcount )
  {
    const char* ps = str;
    const char* fs;
    
    zap();
    
    int rl = strlen( res );
    
    String s;
    while( (fs = strstr( ps, res )) )
      {
      if ( maxcount != -1 )
        {
        maxcount--;
        if ( maxcount == 0 ) break;
        }
      int l = fs - ps;
      s.setn( ps, l );
      push( s );
      ps = s + rl;
      }
    push( ps );
  };

  const char* VArray::join( const char* glue )
  {
    _ret_str = "";
    for( int z = 0; z < _count-1; z++ )
      {
      _ret_str += get( z );
      _ret_str += glue;
      }
    _ret_str += get( _count-1 );
    return _ret_str.data();
  };

  void VArray::print()
  {
    int z;
    for( z = 0; z < count(); z++ )
      printf( "%d=%s\n", z, get(z) );
  }

/***************************************************************************
**
** VTRIE
**
****************************************************************************/

  VTrie::VTrie()
  {
    _count = 0;
    _depth = 0;
    _nodes = 0;
    root = new VTrieNode();
    _nodes++;
  }

  VTrie::VTrie( const VArray& arr )
  {
    _count = 0;
    _depth = 0;
    _nodes = 0;
    root = new VTrieNode();
    _nodes++;
    *this = arr;
  };

  VTrie::VTrie( const VTrie& tr )
  {
    _count = 0;
    _depth = 0;
    _nodes = 0;
    root = new VTrieNode();
    _nodes++;
    *this = tr;
  };


  VTrie::~VTrie()
  {
    zap_node( root );
  }

  void VTrie::zap_node( VTrieNode *node )
  {
    if ( node->next ) zap_node( node->next );
    if ( node->down ) zap_node( node->down );
    if ( node->data ) delete node->data;
    delete node;
    _nodes--;
  }

  void VTrie::del_node( VTrieNode *parent, VTrieNode *node, const char* key )
  {
    ASSERT( key[0] );
    if ( key[1] == 0 )
      {
      if ( key[0] == node->c )
        {
        if ( node->data )
          delete node->data;
        node->data = NULL;
        }
      else
        return; /* not found */
      }
    else
      {
      if ( key[0] == node->c )
        {
        if ( node->down )
          del_node( node, node->down, key+1 );
        else
          return; /* not found */
        }
      else
        {
        if ( node->next )
          del_node( node, node->next, key );
        else
          return; /* not found */
        }
      }

      if ( node->data ) return;
      if ( node->down ) return;
      if ( node->next )
        {
        if ( parent->down == node )
          parent->down = node->next;
        else
          parent->next = node->next;
        }
      else
        {
        if ( parent->down == node )
          parent->down = NULL;
        else
          parent->next = NULL;
        }
      delete node;
      _nodes--;
  }

  void VTrie::key_node( VTrieNode *node,  VArray* arr )
  {
    int kl = str_len( temp_key );
    if ( node->c ) str_add_ch( temp_key, node->c );
    if ( node->data ) arr->push( temp_key );
    if ( node->down ) key_node( node->down, arr );
    str_sleft( temp_key, kl );
    if ( node->next ) key_node( node->next, arr );
  };

  void VTrie::value_node( VTrieNode *node,  VArray* arr )
  {
    if ( node->data ) arr->push( node->data->data() );
    if ( node->down ) value_node( node->down, arr );
    if ( node->next ) value_node( node->next, arr );
  };

  VTrieNode* VTrie::find_node( const char* key )
  {
    if ( !key || !key[0] ) return NULL;

    VTrieNode *current = root;
    const char* keyp = key;
    while( *keyp )
      {
      current = current->down;
      if ( !current )
        {
        return NULL;
        }
      else
        {
        while( current && current->c != *keyp )
          {
          current = current->next;
          }
        if ( !current )
          {
          return NULL;
          }
        }
      keyp++;
      }
    if ( current )
      {
      if ( ! current->data ) current->data = new String();
      return current;
      }
    return NULL;
  };

  int VTrie::set( const char* key, const char* data )
  {
    if ( !key || !key[0] ) return _count;
    int sl = strlen( key );
    if ( sl > _depth )
      _depth = sl;

    VTrieNode *current = root;
    const char* keyp = key;
    while( *keyp )
      {
      VTrieNode *parent = current;
      current = current->down;
      if ( !current )
        {
        current = new VTrieNode();
        parent->down = current;
        current->c = *keyp;
        _nodes++;
        }
      else
        {
        while( current && current->c != *keyp )
          {
          parent = current;
          current = current->next;
          }
        if ( !current )
          {
          current = new VTrieNode();
          parent->next = current;
          current->c = *keyp;
          _nodes++;
          }
        }
      keyp++;
      }
    if ( current->data )
      delete current->data;
    else
      _count++;
    current->data = new String;
    current->data->set( data );
    return _count;
  };

  const char* VTrie::get( const char* key )
  {
    VTrieNode* node = find_node( key );
    if ( node )
      return node->data->data();
    else
      return NULL;
  };

  void VTrie::del( const char* key )
  {
    if ( !key || !key[0] || !root->down ) return;
    del_node( root, root->down, key );
  };

  void VTrie::zap()
  {
    zap_node( root );
    root = new VTrieNode();
    _nodes++;
  };

  int VTrie::keys( VArray* arr )
  {
    arr->zap();
    temp_key = "";
    key_node( root, arr );
    return arr->count();
  };

  int VTrie::values( VArray* arr )
  {
    arr->zap();
    temp_key = "";
    value_node( root, arr );
    return arr->count();
  };

  int VTrie::merge( VTrie *tr )
  {
    VArray arr;
    int cnt = tr->keys( &arr );
    for( int z = 0; z < cnt; z++ )
      set( arr[z], tr->get( arr[z] ) );
    return _count;
  };

  int VTrie::merge( VArray *arr )
  {
    while( arr->get( 0 ) )
      {
      if ( arr->get( 1 ) )
        {
        set( arr->get( 0 ), arr->get( 1 ) );
        arr->del( 0 );
        arr->del( 0 );
        }
      else
        {
        set( arr->get( 0 ), "" );
        arr->del( 0 );
        }
      }
    return _count;
  };

  int VTrie::fload( const char* fname )
  {
    FILE* f = fopen( fname, "rt" );
    if (!f) return 1;
    int r = fload( f );
    fclose(f);
    return r;
  };

  int VTrie::fsave( const char* fname )
  {
    FILE* f = fopen( fname, "wt" );
    if (!f) return 1;
    int r = fsave( f );
    fclose(f);
    return r;
  };

  int VTrie::fload( FILE* f )
  {
    VArray arr;
    int r = arr.fload( f );
    if ( r == 0 )
      merge( &arr );
    return r;
  };

  int VTrie::fsave( FILE* f )
  {
    VArray arr;
    arr.merge( this );
    return arr.fsave( f );
  };

  void VTrie::print_node( VTrieNode* node )
  {
    if (!node) return;
    printf( "---------------------------------\n" );
    printf( "this = %p\n", node );
    printf( "key  = %c\n", node->c );
    printf( "next = %p\n", node->next );
    printf( "down = %p\n", node->down );
    printf( "data = %s\n", node->data->data() );
    print_node( node->next );
    print_node( node->down );
  };

  void VTrie::print()
  {
    VArray va;
    keys( &va );
    int z;
    for( z = 0; z < va.count(); z++ )
      printf( "%s=%s\n", va[z].data(), get( va[z] ) );
  }

/***************************************************************************
**
** VREGEXP
**
****************************************************************************/

  const char* VRegexp::sub( int n )
  {
    if(re->startp[n] != NULL && re->endp[n] != NULL)
      {
      int i = (int)(re->endp[n] - re->startp[n]);
      substr.setn(re->startp[n],i);
      return substr.data();
      }
    else
      {
      substr = "";
      return NULL;
      }
  };


